# -*- coding: utf-8 -*-
# Part of Odoo. See LICENSE file for full copyright and licensing details.

from odoo.exceptions import AccessDenied

from odoo import api, models, fields, registry, SUPERUSER_ID


class TierValidation(models.AbstractModel):
    _inherit = "tier.validation"
    _to_approve_state = "to_approve"
    _approved_state = "approved"
    _rejected_state = 'rejected'
    _subject_prefix = 'subject_prefix'
    _subject_code = 'code'
    _enable_restart_validation = True

    enable_restart_validation = fields.Boolean(
        compute="_compute_enable_restart_validation"
    )

    def _compute_enable_restart_validation(self):
        for rec in self:
            rec.enable_restart_validation = self._enable_restart_validation

    # get list user who already taken action on this request
    # user_approveds = fields.Many2many("res.users", "tier_validation_user_approveds_rel", compute="_compute_list_approved", store=True)

    # @api.depends("review_ids", "review_ids.status")
    # def _compute_list_approved(self):
    #     for record in self:
    #         list_approval = (
    #             self.env["tier.review"]
    #             .search(
    #                 [
    #                     ("model", "=", self._name),
    #                     ("res_id", "=", record.id),
    #                     ("status", "!=", "pending"),
    #                     ("done_by", "!=", False),
    #                 ],
    #             )
    #             .done_by
    #         )
    #         record.user_approveds = list_approval

    def request_validation(self):
        created_trs = super(TierValidation, self).request_validation()
        for rec in self:
            rec.update({
                self._state_field: self._to_approve_state
            })
        return created_trs

    def is_all_reviews_approve(self):
        review_status = self.review_ids.mapped('status')
        if 'pending' in review_status or 'rejected' in review_status:
            return False
        return True

    def is_reviews_reject(self):
        review_status = self.review_ids.mapped('status')
        if 'rejected' in review_status:
            return True
        return False

    def _validate_tier(self, tiers=None):
        self.ensure_one()
        current_approvers = self.get_current_reviewer()
        tier_reviews = tiers or self.review_ids
        user_reviews = tier_reviews.filtered(
            lambda r: r.status == "pending" and (self.env.user in r.reviewer_ids)
        )
        user_reviews.update(
            {
                "status": "approved",
                "done_by": self.env.user.id,
                "reviewed_date": fields.Datetime.now(),
            }
        )

        if self.is_all_reviews_approve():
            self.update({
                self._state_field: self._approved_state
            })

        for approver in current_approvers:
            if approver.definition_id.notify_status_on_action:

                subscribe = "message_post"
                getattr(self, subscribe)(
                    partner_ids=self.get_approved_reviewer().mapped("reviewer_ids").mapped("partner_id").ids,
                    subtype_xmlid=self._get_accepted_notification_subtype(),
                    body=self._notify_accepted_reviews_body(),
                    subject=f'{self._subject_prefix} Request {str(getattr(self, self._subject_code, self.id))}" was accepted'
                )
            else:
                for review in user_reviews:
                    rec = self.env[review.model].browse(review.res_id)
                    rec._notify_accepted_reviews()

        for approver in current_approvers:
            if approver.definition_id.notify_status_on_next:
                next_approvers = self.get_next_reviewer().mapped("reviewer_ids")
                post = "message_post"
                getattr(self, post)(
                    subtype_xmlid=self._get_requested_notification_subtype(),
                    body=self._notify_requested_review_body(),
                    partner_ids=next_approvers.mapped("partner_id").ids,
                    subject=f'{self._subject_prefix} Required approval for request ID = {str(getattr(self, self._subject_code, self.id))}'
                )

    def _rejected_tier(self, reviews=False):
        self.ensure_one()
        tier_reviews = reviews or self.review_ids
        user_reviews = tier_reviews.filtered(
            lambda r: r.status == "pending" and (self.env.user in r.reviewer_ids)
        )
        user_reviews.write(
            {
                "status": "rejected",
                "done_by": self.env.user.id,
                "reviewed_date": fields.Datetime.now(),
            }
        )

        if self.is_reviews_reject():
            self.with_context({'skip_validation_check': True}).update({
                self._state_field: self._rejected_state
            })

        for approver in self.get_approved_reviewer():
            if approver.definition_id.notify_status_on_action:
                post = "message_post"
                getattr(self.sudo(), post)(
                    subtype_xmlid=self._get_rejected_notification_subtype(),
                    body=self._notify_rejected_review_body(),
                    subject=f'{self._subject_prefix} Request {str(getattr(self, self._subject_code, self.id))}" was rejected'
                )

    def get_current_reviewer(self):
        self.ensure_one()
        sequences = self._get_sequences_to_approve(self.env.user)
        reviews = self.review_ids.filtered(lambda l: l.sequence in sequences)
        user_reviews = reviews.filtered(
            lambda r: r.status == "pending" and (self.env.user in r.reviewer_ids)
        )
        return user_reviews

    def get_approved_reviewer(self):
        self.ensure_one()
        reviews = self.review_ids.filtered(lambda l: l.status == 'approved')
        user_reviews = reviews.filtered(
            lambda r: r.status == 'approved'
        )
        return user_reviews

    def get_next_reviewer(self):
        self.ensure_one()
        sequences = self._get_next_sequences_to_approve()
        reviews = self.review_ids.filtered(lambda l: l.sequence == sequences)
        return reviews

    def _get_next_sequences_to_approve(self):
        all_reviews = self.review_ids.filtered(lambda r: r.status == "pending")
        pending_sequence = all_reviews.mapped("sequence")
        min_seq = min(pending_sequence)
        return min_seq

    def _notify_rejected_review(self):
        post = "message_post"
        if hasattr(self, post):
            # Notify state change
            getattr(self.sudo(), post)(
                subtype_xmlid=self._get_rejected_notification_subtype(),
                body=self._notify_rejected_review_body(),
                subject=f'{self._subject_prefix} Request {str(getattr(self, self._subject_code, self.id))}" was rejected'
            )

    def _notify_review_requested(self, tier_reviews):
        subscribe = "message_subscribe"
        post = "message_post"
        if hasattr(self, post) and hasattr(self, subscribe):
            for rec in self:
                users_to_notify = tier_reviews.filtered(
                    lambda r: r.definition_id.notify_on_create and r.res_id == rec.id
                ).mapped("reviewer_ids")
                # Subscribe reviewers and notify
                getattr(rec, subscribe)(
                    partner_ids=users_to_notify.mapped("partner_id").ids
                )
                getattr(rec, post)(
                    subtype_xmlid=self._get_requested_notification_subtype(),
                    body=rec._notify_requested_review_body(),
                    subject=f'{self._subject_prefix} Required approval for request ID = {str(getattr(self, self._subject_code, self.id))}'
                )

    def _notify_restarted_review(self):
        post = "message_post"
        if hasattr(self, post):
            getattr(self.sudo(), post)(
                subtype_xmlid=self._get_restarted_notification_subtype(),
                body=self._notify_restarted_review_body(),
                subject=f'{self._subject_prefix} Request {str(getattr(self, self._subject_code, self.id))}" was reset'
            )

    def _notify_accepted_reviews(self):
        post = "message_post"
        if hasattr(self, post):
            # Notify state change
            getattr(self.sudo(), post)(
                subtype_xmlid=self._get_accepted_notification_subtype(),
                body=self._notify_accepted_reviews_body(),
                subject=f'{self._subject_prefix} Request {str(getattr(self, self._subject_code, self.id))}" was accepted'
            )
